import time
import usb
import sys
import struct
import array
import random

# Definitions of commands sent in USB packets.

READ_VERSION_CMD       = 0x00  # Read the product version information.
READ_FLASH_CMD         = 0x01  # Read from the device flash.
WRITE_FLASH_CMD        = 0x02  # Write to the device flash.
ERASE_FLASH_CMD        = 0x03  # Erase the device flash.
READ_EEDATA_CMD        = 0x04  # Read from the device EEPROM.
WRITE_EEDATA_CMD       = 0x05  # Write to the device EEPROM.
READ_CONFIG_CMD        = 0x06  # Read from the device configuration memory.
WRITE_CONFIG_CMD       = 0x07  # Write to the device configuration memory.
ID_BOARD_CMD           = 0x31  # Flash the device LED to identify which device is being communicated with.
UPDATE_LED_CMD         = 0x32  # Change the state of the device LED.
INFO_CMD               = 0x40  # Get information about the USB interface.
SENSE_INVERTERS_CMD    = 0x41  # Sense inverters on TCK and TDO pins of the secondary JTAG port.
TMS_TDI_CMD            = 0x42  # Send a single TMS and TDI bit.
TMS_TDI_TDO_CMD        = 0x43  # Send a single TMS and TDI bit and receive TDO bit.
TDI_TDO_CMD            = 0x44  # Send multiple TDI bits and receive multiple TDO bits.
TDO_CMD                = 0x45  # Receive multiple TDO bits.
TDI_CMD                = 0x46  # Send multiple TDI bits.
RUNTEST_CMD            = 0x47  # Pulse TCK a given number of times.
NULL_TDI_CMD           = 0x48  # Send string of TDI bits.
PROG_CMD               = 0x49  # Change the level of the FPGA PROGRAM# pin.
SINGLE_TEST_VECTOR_CMD = 0x4a  # Send a single, byte-wide test vector.
GET_TEST_VECTOR_CMD    = 0x4b  # Read the current test vector being output.
SET_OSC_FREQ_CMD       = 0x4c  # Set the frequency of the DS1075 oscillator.
ENABLE_RETURN_CMD      = 0x4d  # Enable return of info in response to a command.
DISABLE_RETURN_CMD     = 0x4e  # Disable return of info in response to a command.
TAP_SEQ_CMD            = 0x4f  # Send multiple TMS & TDI bits while receiving multiple TDO bits.
FLASH_ONOFF_CMD        = 0x50  # Enable/disable the FPGA configuration flash.
RESET_CMD              = 0xff  # Cause a power-on reset.

def bytes(*b):
    return array.array('B', b).tostring()

from jtag import Jtag
from bitstream import Bitstream, BitstreamHex, BitFile

reverse_bits = [
    0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0, 0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
    0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8, 0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
    0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4, 0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
    0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec, 0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
    0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2, 0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
    0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea, 0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
    0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6, 0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
    0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee, 0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
    0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1, 0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
    0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9, 0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
    0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5, 0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
    0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed, 0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
    0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3, 0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
    0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb, 0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
    0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7, 0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
    0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef, 0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff,
]

class XuLA(Jtag):

    # see ug332, Table 9-5 p 207:
    # Spartan-3 Boundary Scan Instructions
    EXTEST       = Bitstream(6, int("001111", 2))
    SAMPLE       = Bitstream(6, int("000001", 2))
    USER1        = Bitstream(6, int("000010", 2))
    USER2        = Bitstream(6, int("000011", 2))
    CFG_OUT      = Bitstream(6, int("000100", 2))
    CFG_IN       = Bitstream(6, int("000101", 2))
    INTEST       = Bitstream(6, int("000111", 2))
    USERCODE     = Bitstream(6, int("001000", 2))
    IDCODE       = Bitstream(6, int("001001", 2))
    HIGHZ        = Bitstream(6, int("001010", 2))
    JPROGRAM     = Bitstream(6, int("001011", 2))
    JSTART       = Bitstream(6, int("001100", 2))
    JSHUTDOWN    = Bitstream(6, int("001101", 2))
    ISC_ENABLE   = Bitstream(6, int("010000", 2))
    ISC_PROGRAM  = Bitstream(6, int("010001", 2))
    ISC_NOOP     = Bitstream(6, int("010100", 2))
    ISC_READ     = Bitstream(6, int("010101", 2))
    ISC_DISABLE  = Bitstream(6, int("010110", 2))
    ISC_DNA      = Bitstream(6, int("110001", 2))
    BYPASS       = Bitstream(6, int("111111", 2))

    def __init__(self):
        buses = usb.busses()
        xula = None
        for bus in buses:
            for device in bus.devices:
                if device.idVendor == 0x04d8 and device.idProduct == 0xff8c:
                    xula = device
        if xula is None:
            print "No XuLA device found on USB bus"
            sys.exit(1)

        print "Found XuLA on USB bus"

        self.handle = xula.open()

        try:
            self.handle.detachKernelDriver(0)
        except usb.USBError, error:
            # print "detachKernelDriver exception", error
            pass
        self.handle.claimInterface(0)

        def powercycle():
            m = bytes(RESET_CMD) + (chr(0) * 31)
            self.handle.bulkWrite(usb.ENDPOINT_OUT + 1, m, 1000)
            time.sleep(4)

        self.handle.resetEndpoint(usb.ENDPOINT_OUT + 1)
        self.handle.resetEndpoint(usb.ENDPOINT_IN + 1)
        self.handle.reset()
        m = bytes(INFO_CMD, 0)
        self.handle.bulkWrite(usb.ENDPOINT_OUT + 1, m, 1000)
        try:
            device_info = self.handle.bulkRead(usb.ENDPOINT_IN + 1, 32, 1000)
        except usb.USBError, error:
            powercycle()

        if (sum(device_info)) & 0xff != 0:
            print 'Device info Checksum error'
            sys.exit(1)
        print '  Product ID: %02x %02x' % device_info[1:3]
        print '  Version: %d %d' % device_info[3:5]
        # Desc is 0-terminated string
        desc = device_info[5:-1]
        desclen = desc.index(0)
        print '  Description:', bytes(*desc[:desclen])


    # Sample TDO, output TMS and TDI values, pulse TCK, and return TDO value.
    def tick(self, tms, tdi):
        mask = 0
        if tms:
            mask |= 0x01
        if tdi:
            mask |= 0x02
        m = bytes(TMS_TDI_TDO_CMD, mask) # + (chr(0) * 30)
        self.handle.bulkWrite(usb.ENDPOINT_OUT + 1, m, 1000)
        r = self.handle.bulkRead(usb.ENDPOINT_IN + 1, 2, 1000)
        return (r[1] & 0x04) != 0;

    def bulktdi(self, bs):
        m = struct.pack("<BI", TDI_CMD, len(bs))
        self.handle.bulkWrite(usb.ENDPOINT_OUT + 1, m, 1000)
        t = time.time()
        if 0:
            b256 = bs # ([0] * (ps*8 - len(bs))) + list(bs)
            w256 = [(b << (i & 7)) for (i, b) in enumerate(b256)]
            m = array.array('B', [sum(w256[i:i+8]) for i in range(0, len(w256), 8)]).tostring()
        else:
            m = array.array('B', bs.tostring())
            m = array.array('B', [reverse_bits[c] for c in m.tolist()]).tostring()
        # print "(Bulk %d)" % len(m), ["%02x" % ord(c) for c in m[:50]]
        # print ["%02x" % ord(x) for x in m]

        self.handle.bulkWrite(usb.ENDPOINT_OUT + 1, m, 1000)
        self.debug_tms(1)
        print "took", time.time() - t

    def word(self, bs):
        m = struct.pack("<BI", TDI_TDO_CMD, len(bs))
        self.handle.bulkWrite(usb.ENDPOINT_OUT + 1, m, 1000)
        b256 = bs # ([0] * (ps*8 - len(bs))) + list(bs)
        w256 = [(b << (i & 7)) for (i, b) in enumerate(b256)]
        m = array.array('B', [sum(w256[i:i+8]) for i in range(0, len(w256), 8)]).tostring()

        self.handle.bulkWrite(usb.ENDPOINT_OUT + 1, m, 1000)
        self.debug_tms(1)
        recv = self.handle.bulkRead(usb.ENDPOINT_IN + 1, 32, 1000)
        return recv[0]

    def bulktms(self, bs):
        GET_TDO_MASK = 0x01                       # Set if gathering TDO bits.
        PUT_TMS_MASK = 0x02                       # Set if TMS bits are included in the packets.
        TMS_VAL_MASK = 0x04                       # Static value for TMS if PUT_TMS_MASK is cleared.
        PUT_TDI_MASK = 0x08                       # Set if TDI bits are included in the packets.
        TDI_VAL_MASK = 0x10                       # Static value for TDI if PUT_TDI_MASK is cleared.
        DO_MULTIPLE_PACKETS_MASK = 0x80           # Set if command extends over multiple USB packets.
        m = struct.pack("<BIB", TAP_SEQ_CMD, len(bs) * 2, PUT_TDI_MASK | PUT_TMS_MASK)
        b256 = bs # ([0] * (ps*8 - len(bs))) + list(bs)
        w256 = [(b << (i & 7)) for (i, b) in enumerate(b256)]
        d = array.array('B', [sum(w256[i:i+8]) for i in range(0, len(w256), 8)]).tostring()

        # cmd, len, flags, tms, tdi
        m += d[0] + chr(0)
        print repr(m)
        self.handle.bulkWrite(usb.ENDPOINT_OUT + 1, m, 1000)
        for g in bs:
            self.debug_tms(g)

    def querychain(self):
        if 0:
            # do not need to count devices, because there is only one
            self.go_states(1,1,1,1,1)
            self.go_states(0,1,1,0,0)
            self.assert_state("Shift-IR")
            self.send_str("1" * 1000, 1)
            self.assert_state("Exit1-IR")
            self.go_states(1,1,0,0)
            self.send_str("0" * 1000, 0)
            ndevices = 0
            while self.do_bit(0, 1) == 0:
                ndevices += 1
        else:
            ndevices = 1

        self.go_states(1,1,1,1,1)
        self.go_states(0,1,0,0)

        return [self.do_nbit_cycle(32, 0) for i in range(ndevices)]


    def progpin(self, v):
        m = bytes(PROG_CMD, v) + (chr(0) * 30)
        self.handle.bulkWrite(usb.ENDPOINT_OUT + 1, m, 1000)

    def idcode(self):
        return self.LoadBSIRthenBSDR(self.IDCODE, Bitstream(32, 0), receive = True)

    def usercode(self):
        return self.LoadBSIRthenBSDR(self.USERCODE, Bitstream(32, 0), receive = True)

    def pulseTCK(self, c):
        self.assert_state("Run-Test/Idle")
        for i in range(c):
            self.do_bit(0, 0)

    def DNA(self):
        self.tlr()
        self.LoadBSIRthenBSDR(self.JPROGRAM, None)
        time.sleep(.5)
        print "ISC_ENABLE", self.LoadBSIRthenBSDR(self.ISC_ENABLE, Bitstream(5, 0), receive = True)
        dna = self.LoadBSIRthenBSDR(self.ISC_DNA, Bitstream(57, 0), receive = True)
        print "ISC_DNA", hex(dna)
        return dna

    def ccl(self, rw, reg, cnt):
        """
        Return the 16-bit hex string for a Configuration
        Control Logic operation, (ug332 page 323).
        """
        assert rw in "rw"
        assert (0 <= reg < 64)
        assert (0 <= cnt < 32)
        m = 1 << 13     # Type 1 packet
        if rw == "r":
            m |= (1 << 11)
        else:
            m |= (2 << 11)
        m |= reg << 5
        m |= cnt
        return "%04x" % m

    def rdccl(self, reg):
        read_status = [
            "aa99",
            "2000",
            self.ccl('r', reg, 1),
            "2000",
            "2000",
        ]
        commands = "".join(read_status)
        self.LoadBSIRthenBSDR(self.CFG_IN, BitstreamHex(commands))
        return self.LoadBSIRthenBSDR(self.CFG_OUT, Bitstream(32, 0), receive = True)

    def resetcrc(self):
        print 'CRC=', hex(self.rdccl(0))

        write_crc = [
            "aa99",
            "2000",
            self.ccl('w', 5, 1),
            "0007",
            "2000",
        ]
        commands = "".join(write_crc)
        print commands
        self.LoadBSIRthenBSDR(self.CFG_IN, BitstreamHex(commands))
        print self.LoadBSIRthenBSDR(self.CFG_OUT, Bitstream(32, 0), receive = True)
        print 'CRC=', hex(self.rdccl(0))

    # see ug332, page 340
    def status(self):
        # packet format see ug332, page 323
        read_status = [
            "aa99",
            "2000",
            self.ccl('r', 8, 1),
            "2000",
            "2000",
        ]
        commands = "".join(read_status)
        self.LoadBSIRthenBSDR(self.CFG_IN, BitstreamHex(commands))
        status = self.LoadBSIRthenBSDR(self.CFG_OUT, Bitstream(32, 0), receive = True)

        print "status = %04x" % status
        vv = [
            ('SYNC_TIMEOUT',    1 & (status >> 15)),
            ('SEU_ERR',         1 & (status >> 14)),
            ('DONE',            1 & (status >> 13)),
            ('INIT',            1 & (status >> 12)),
            ('MODE',            7 & (status >> 9)),
            ('VSEL',            7 & (status >> 6)),
            ('GHIGH_B',         1 & (status >> 5)),
            ('GWE',             1 & (status >> 4)),
            ('GTS_CFG_B',       1 & (status >> 3)),
            ('DCM_LOCK',        1 & (status >> 2)),
            ('ID_ERROR',        1 & (status >> 1)),
            ('CRC_ERROR',       1 & (status >> 0))]
        for nm,v in vv:
            print "%16s %d" % (nm, v)

        return status

    # xapp139 - 
    # http://www.xilinx.com/support/documentation/application_notes/xapp452.pdf
    def load(self, bs):
        # Must follow JPROGRAM with CFG_IN to keep device locked
        # to JTAG.  See AR 16829.
        self.LoadBSIRthenBSDR(self.JPROGRAM, None)
        self.LoadBSIRthenBSDR(self.CFG_IN, None)
        # print list(bs)[256:512]
        time.sleep(0.001)
        self.LoadBSIRthenBSDR(self.CFG_IN, bs)
        self.tlr()
        self.LoadBSIRthenBSDR(self.JSTART, None)
        return
        
        S3_BYPASS		= Bitstream(6, int("111111", 2))
        S3_JPROGRAM		= Bitstream(6, int("001011", 2))
        S3_CFG_IN		= Bitstream(6, int("000101", 2))
        S3_JSTART		= Bitstream(6, int("001100", 2))
        S3_JSHUTDOWN	= Bitstream(6, int("001101", 2))

        self.LoadBSIRthenBSDR(S3_JPROGRAM, None)
        self.pulseTCK(1000)

        self.LoadBSIRthenBSDR(S3_CFG_IN, Bitstream(192, 0x0000000000000000e00000008001000c66aa9955ffffffff))

        self.LoadBSIRthenBSDR(S3_JSHUTDOWN, None)
        self.pulseTCK(12)

        self.tlr()
        self.LoadBSIRthenBSDR(S3_CFG_IN, Bitstream(96, 0x000000000000000066aa9955))
        self.LoadBSIRthenBSDR(S3_CFG_IN, bs)

        for j in range(2):
            for i in range(2):
                self.tlr()
                self.go_state(0)
                self.assert_state("Run-Test/Idle");
            self.LoadBSIRthenBSDR(S3_JSTART, None)
            self.pulseTCK(12)
            self.LoadBSIRthenBSDR(S3_BYPASS, None)
            self.LoadBSIRthenBSDR(S3_BYPASS, None)
        self.tlr()
        self.go_state(0)
        self.assert_state("Run-Test/Idle");

